# Copyright 2022 University of Waterloo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# from __future__ import annotations

from datetime import date
from typing import List, Optional

from pydantic import BaseModel, Field, validator

from almonaut import common_validators


class Library1(BaseModel):
    # used by class ElectronicCollection
    value: str


class Type2(BaseModel):
    # used by class ElectronicCollection
    value: str
    desc: str


class Interface1(BaseModel):
    # used by class ElectronicCollection
    name: str


class IsSelective(BaseModel):
    value: str
    desc: Optional[str] = None


class AccessType1(BaseModel):
    # used by class ElectronicCollection
    value: str


class CounterPlatform1(BaseModel):
    # used by class ElectronicCollection
    value: str
    desc: Optional[str] = None


class PoLine1(BaseModel):
    # used by class ElectronicCollection
    value: str


class License1(BaseModel):
    # used by class ElectronicCollection
    value: str


class Number1(BaseModel):
    # used by class ElectronicCollection
    value: str


class Free2(BaseModel):
    value: str


class ProxyEnabled3(BaseModel):
    # used by class ElectronicCollection
    value: str


class Language(BaseModel):
    value: str


class MmsId1(BaseModel):
    # used by class ResourceMetadata1
    value: str


class ResourceMetadata1(BaseModel):
    # used by class ElectronicCollection
    mms_id: MmsId1


class Note2(BaseModel):
    # used by class ElectronicCollection
    content: str
    creation_date: date
    created_by: str
    type_: str = Field(..., alias='type')

    _strip_gmt_date_z = validator('creation_date', pre=True, allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class Group2(BaseModel):
    # used by class GroupSetting2
    value: str


class ProxyEnabled4(BaseModel):
    # used by class GroupSetting2
    value: str


class Portfolios2(BaseModel):
    # used by class ElectronicCollection
    value: int
    link: Optional[str] = None


class GroupSetting2(BaseModel):
    group: Group2
    proxy_enabled: ProxyEnabled4
    proxy: str
    public_name: str
    authentication_note: str
    public_note: str


class CdiUpdateFrequency(BaseModel):
    value: str
    desc: Optional[str] = None


class SearchRightsInCdi(BaseModel):
    value: str
    desc: str


class FullTextLinkingInCdi(BaseModel):
    value: str
    desc: str


class CdiNewspapersSearch(BaseModel):
    value: str
    desc: str


class FullTextRightsInCdi(BaseModel):
    value: str
    desc: str


class CdiType(BaseModel):
    value: str
    desc: str


class CdiInfo(BaseModel):
    cdi_collection_id: Optional[str] = None
    number_of_records_in_cdi: Optional[int] = None
    cdi_update_frequency: Optional[CdiUpdateFrequency] = None
    search_rights_in_cdi: Optional[SearchRightsInCdi] = None
    full_text_linking_in_cdi: Optional[FullTextLinkingInCdi] = None
    cdi_newspapers_search: Optional[CdiNewspapersSearch] = None
    provider_coverage: Optional[bool] = None
    resource_types: Optional[str] = None
    full_text_rights_in_cdi: Optional[FullTextRightsInCdi] = None
    cdi_type: Optional[CdiType] = None
    coverage_percentage: Optional[str] = None


class ElectronicCollection(BaseModel):
    id_: str = Field(..., alias='id')
    is_local = bool
    link: str
    public_name: str
    description: str
    description_override: Optional[str] = None
    internal_description: Optional[str] = None
    library: Optional[Library1] = None
    type_: Type2 = Field(..., alias='type')
    interface: Optional[Interface1] = None
    is_selective: Optional[IsSelective] = None
    access_type: Optional[AccessType1] = None
    counter_platform: CounterPlatform1
    po_line: Optional[PoLine1] = None
    activation_date: Optional[date] = None
    expected_activation_date: Optional[date] = None
    license_: Optional[License1] = Field(..., alias='license')
    number: Optional[List[Number1]] = None
    alternative_title: Optional[str] = None
    source: Optional[str] = None
    creator: Optional[str] = None
    url: Optional[str] = None
    url_override: Optional[str] = None
    free: Optional[Free2] = None
    proxy_enabled: Optional[ProxyEnabled3] = None
    proxy: Optional[str] = None
    language: Optional[Language] = None
    category: Optional[str] = None
    resource_metadata: Optional[ResourceMetadata1] = None
    authentication_note: Optional[str] = None
    public_note: Optional[str] = None
    notes: Optional[List[Note2]] = None
    group_settings: Optional[List[GroupSetting2]] = None
    portfolios: Portfolios2
    do_not_show_as_full_text_available_in_cdi_even_if_active_in_alma: Optional[bool] = None
    cdi_search_activation_status: Optional[bool] = None
    cdi_only_full_text_activation: Optional[bool] = None
    cdi_local_notes: Optional[str] = None
    cdi_info: CdiInfo

    _strip_gmt_date_z = validator('activation_date', 'expected_activation_date', pre=True, allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class ElectronicCollections(BaseModel):
    total_record_count: int
    electronic_collections: List[ElectronicCollection] = Field(..., alias='electronic_collection')


class Type1(BaseModel):
    # used by class ElectronicService
    value: str
    desc: str


class ActivationStatus(BaseModel):
    value: str
    desc: str


class ServiceTemporarilyUnavailable(BaseModel):
    value: str


class LinkResolverPlugin(BaseModel):
    value: str


class UrlType1(BaseModel):
    # used by class ElectronicService
    value: str


class UrlTypeOverride1(BaseModel):
    # used by class ElectronicService
    value: str


class Free1(BaseModel):
    # used by class ElectronicService
    value: str


class CrossrefSupported(BaseModel):
    value: str


class CrossrefEnabled(BaseModel):
    value: str


class ProxyEnabled1(BaseModel):
    # used by class ElectronicService
    value: str


class LinkingLevel(BaseModel):
    value: str


class Note1(BaseModel):
    # used by class ElectronicService
    content: str
    creation_date: date
    created_by: str
    type_: str = Field(..., alias='type')

    _strip_gmt_date_z = validator('creation_date', pre=True, allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class Group1(BaseModel):
    # used by class GroupSetting1
    value: str


class Portfolios1(BaseModel):
    # used by class ElectronicService
    value: int
    link: str


class ProxyEnabled2(BaseModel):
    # used by class GroupSetting1
    value: str


class GroupSetting1(BaseModel):
    # used by class ElectronicService
    group: Group1
    proxy_enabled: ProxyEnabled2
    proxy: str
    public_name: str
    authentication_note: str
    public_note: str


class ElectronicService(BaseModel):
    id_: str = Field(..., alias='id')
    link: str
    is_local: bool
    type_: Type1 = Field(..., alias='type')
    public_description: str
    internal_description: Optional[str] = None
    public_description_override: str
    activation_status: ActivationStatus
    activate_new_portfolios: Optional[bool] = None
    active_from_date: Optional[date] = None
    active_until_date: Optional[date] = None
    service_temporarily_unavailable: Optional[ServiceTemporarilyUnavailable] = None
    service_unavailable_date: Optional[date] = None
    service_unavailable_reason: Optional[str] = None
    parser: Optional[str] = None
    parser_override: Optional[str] = None
    parser_parameters: Optional[str] = None
    parser_parameters_override: Optional[str] = None
    link_resolver_plugin: Optional[LinkResolverPlugin] = None
    url_type: Optional[UrlType1] = None
    url_type_override: Optional[UrlTypeOverride1] = None
    dynamic_url: Optional[str] = None
    dynamic_url_override: Optional[str] = None
    free: Optional[Free1] = None
    crossref_supported: Optional[CrossrefSupported] = None
    crossref_enabled: Optional[CrossrefEnabled] = None
    proxy_enabled: Optional[ProxyEnabled1] = None
    proxy: Optional[str] = None
    linking_level: Optional[LinkingLevel] = None
    authentication_note: Optional[str] = None
    public_note: Optional[str] = None
    notes: Optional[List[Note1]] = None
    group_settings: Optional[List[GroupSetting1]] = None
    portfolios: Portfolios1

    _strip_gmt_date_z = validator('active_from_date', 'active_until_date',
                                  'service_unavailable_date', pre=True,
                                  allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class ElectronicServices(BaseModel):
    # Alma API total_record_count for this API can be null
    total_record_count: Optional[int] = None
    electronic_services: List[ElectronicService] = Field(..., alias='electronic_service')


class MmsId2(BaseModel):
    # used by class ResourceMetadata2
    link: str
    value: str


class ResourceMetadata2(BaseModel):
    # used by class Portfolio
    mms_id: MmsId2
    title: str
    author: Optional[str] = None
    issn: Optional[str] = None


class Availability(BaseModel):
    value: str
    desc: str


class MaterialType(BaseModel):
    value: str


class Library2(BaseModel):
    # used by class Portfolio
    value: str


class AccessType2(BaseModel):
    # used by class Portfolio
    value: str


class CounterPlatform2(BaseModel):
    # used by class Portfolio
    value: str


class UrlType2(BaseModel):
    # used by class LinkingDetails
    value: str


class UrlTypeOverride2(BaseModel):
    # used by class LinkingDetails
    value: str


class ProxyEnabled5(BaseModel):
    # used by class LinkingDetails
    value: str


class LinkingDetails(BaseModel):
    url: str
    url_type: UrlType2
    url_type_override: UrlTypeOverride2
    dynamic_url: str
    dynamic_url_override: str
    static_url: str
    static_url_override: str
    parser_parameters: str
    parser_parameters_override: str
    proxy_enabled: ProxyEnabled5
    proxy: str


class CoverageInUse(BaseModel):
    value: str


class FromMonth1(BaseModel):
    # used by class GlobalDateCoverageParameter
    value: str


class FromDay1(BaseModel):
    # used by class GlobalDateCoverageParameter
    value: str


class UntilMonth1(BaseModel):
    # used by class GlobalDateCoverageParameter
    value: str


class UntilDay1(BaseModel):
    # used by class GlobalDateCoverageParameter
    value: str


class GlobalDateCoverageParameter(BaseModel):
    from_year: str
    from_month: FromMonth1
    from_day: FromDay1
    from_volume: str
    from_issue: str
    until_year: str
    until_month: UntilMonth1
    until_day: UntilDay1
    until_volume: str
    until_issue: str


class FromMonth2(BaseModel):
    # used by class LocalDateCoverageParameter
    value: str


class FromDay2(BaseModel):
    # used by class LocalDateCoverageParameter
    value: str


class UntilMonth2(BaseModel):
    # used by class LocalDateCoverageParameter
    value: str


class UntilDay2(BaseModel):
    # used by class LocalDateCoverageParameter
    value: str


class LocalDateCoverageParameter(BaseModel):
    from_year: str
    from_month: FromMonth2
    from_day: FromDay2
    from_volume: str
    from_issue: str
    until_year: str
    until_month: UntilMonth2
    until_day: UntilDay2
    until_volume: str
    until_issue: str


class FromMonth3(BaseModel):
    # used by class PerpetualDateCoverageParameter
    value: str


class FromDay3(BaseModel):
    # used by class PerpetualDateCoverageParameter
    value: str


class UntilMonth3(BaseModel):
    # used by class PerpetualDateCoverageParameter
    value: str


class UntilDay3(BaseModel):
    # used by class PerpetualDateCoverageParameter
    value: str


class PerpetualDateCoverageParameter(BaseModel):
    from_year: str
    from_month: FromMonth3
    from_day: FromDay3
    from_volume: str
    from_issue: str
    until_year: str
    until_month: UntilMonth3
    until_day: UntilDay3
    until_volume: str
    until_issue: str


class EmbargoOperator1(BaseModel):
    # used by class GlobalEmbargoInformation
    value: str


class GlobalEmbargoInformation(BaseModel):
    embargo_operator: EmbargoOperator1
    number_of_years: str
    number_of_months: str


class EmbargoOperator2(BaseModel):
    # used by class LocalEmbargoInformation
    value: str


class LocalEmbargoInformation(BaseModel):
    embargo_operator: EmbargoOperator2
    number_of_years: str
    number_of_months: str


class EmbargoOperator3(BaseModel):
    # used by class PerpetualEmbargoInformation
    value: str


class PerpetualEmbargoInformation(BaseModel):
    embargo_operator: EmbargoOperator3
    number_of_years: str
    number_of_months: str


class CoverageDetails(BaseModel):
    coverage_in_use: CoverageInUse
    global_date_coverage_parameters: List[GlobalDateCoverageParameter]
    local_date_coverage_parameters: List[LocalDateCoverageParameter]
    perpetual_date_coverage_parameters: List[PerpetualDateCoverageParameter]
    global_embargo_information: GlobalEmbargoInformation
    local_embargo_information: LocalEmbargoInformation
    perpetual_embargo_information: PerpetualEmbargoInformation


class PoLine2(BaseModel):
    # used by class Portfolio
    value: str


class PublicAccessModel(BaseModel):
    value: str


class License2(BaseModel):
    # used by class Portfolio
    value: str


class Vendor(BaseModel):
    link: str
    value: str


class Interface2(BaseModel):
    # used by class Portfolio
    name: str
    vendor: Vendor


class Pda(BaseModel):
    value: str


class Number2(BaseModel):
    # used by class Portfolio
    link: str
    value: str


class Note3(BaseModel):
    # used by class Portfolio
    content: str
    creation_date: date
    created_by: str
    type_: str = Field(..., alias='type')

    _strip_gmt_date_z = validator('creation_date', pre=True, allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class Group3(BaseModel):
    # used by class GroupSetting3
    value: str


class ProxyEnabled6(BaseModel):
    # used by class GroupSetting1
    value: str


class GroupSetting3(BaseModel):
    # used by class Portfolio
    group: Group3
    proxy_enabled: ProxyEnabled6
    proxy: str
    public_name: str
    authentication_note: str
    public_note: str


class Portfolio(BaseModel):
    id_: str = Field(..., alias='id')
    link: str
    is_standalone: Optional[bool] = None
    resource_metadata: ResourceMetadata2
    electronic_collection: Optional[ElectronicCollection] = None
    availability: Availability
    material_type: Optional[MaterialType] = None
    activation_date: Optional[date] = None
    expected_activation_date: Optional[date] = None
    library: Optional[Library2] = None
    access_type: Optional[AccessType2] = None
    counter_platform: Optional[CounterPlatform2] = None
    linking_details: Optional[LinkingDetails] = None
    coverage_details: Optional[CoverageDetails] = None
    po_line: Optional[PoLine2] = None
    public_access_model: Optional[PublicAccessModel] = None
    license_: Optional[License2] = Field(..., alias='license')
    interface: Optional[Interface2] = None
    pda: Optional[Pda] = None
    number: Optional[List[Number2]] = None
    authentication_note: Optional[str] = None
    public_note: Optional[str] = None
    internal_description: Optional[str] = None
    notes: Optional[List[Note3]] = None
    group_settings: Optional[List[GroupSetting3]] = None

    _strip_gmt_date_z = validator('activation_date', 'expected_activation_date', pre=True, allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class Portfolios(BaseModel):
    total_record_count: Optional[int] = None
    portfolios: List[Portfolio] = Field(..., alias='portfolio')



# Copyright 2022 University of Waterloo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests


class AlmaApiSession(requests.Session):
    """The persistent TCP session."""

    def __init__(self, *args, **kwargs):
        """Init method."""
        super(AlmaApiSession, self).__init__(*args, **kwargs)

        self.headers.update({
            'Accept-Charset': 'utf-8',
            'Content-Type': 'text/plain',
        })



# Copyright 2022 University of Waterloo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# from __future__ import annotations

from datetime import date
from typing import List, Optional

from pydantic import BaseModel, Field, validator

from almonaut import common_validators


class AdditionalCharges(BaseModel):
    use_pro_rata: bool
    shipment: int
    overhead: int
    insurance: int
    discount: int


class Administrator(BaseModel):
    email: str
    primary_id: str
    first_name: str
    last_name: str
    link: str


class AcquisitionMethod(BaseModel):
    value: str


class AlternativeCallNumberType(BaseModel):
    value: str


class AvailableForLibrary(BaseModel):
    value: str
    desc: str


class BreakIndicator(BaseModel):
    value: str


class Currency1(BaseModel):
    # used by class Fund
    value: str


class Currency3(BaseModel):
    # used by class Invoice
    value: str


class EntityType(BaseModel):
    value: str
    desc: str


class FiscalPeriod(BaseModel):
    value: str
    desc: str


class ForeignCurrency(BaseModel):
    value: str


class FundCode1(BaseModel):
    # used by class FundDistributionItem
    value: str
    desc: Optional[str] = None


class FundCode2(BaseModel):
    # used by class
    value: str


class InterestedUser(BaseModel):
    primary_id: str
    notify_receiving_activation: bool
    hold_item: bool
    notify_renewal: bool
    notify_cancel: bool


class ItemPolicy(BaseModel):
    value: str


class Library(BaseModel):
    value: str


class License2(BaseModel):
    # used by class PoLine
    value: str
    desc: str


class LicensingAgent(BaseModel):
    value: Optional[str] = None
    desc: Optional[str] = None


class Licensor(BaseModel):
    value: Optional[str] = None
    desc: Optional[str] = None


class Location2(BaseModel):
    # used by class License
    value: str
    desc: Optional[str] = None


class MaterialType(BaseModel):
    value: str
    desc: str


class MmsId(BaseModel):
    value: str


class Note1(BaseModel):
    # used by class PoLine
    note_text: str


class Note2(BaseModel):
    # used by class Invoice
    note_text: Optional[str] = None


class OverencumbranceAllowed(BaseModel):
    value: str
    desc: str


class OverexpenditureAllowed(BaseModel):
    value: str
    desc: str


class Owner1(BaseModel):
    # used by class Fund
    value: str
    desc: str


class Owner2(BaseModel):
    # used by class PoLine
    value: str


class Owner3(BaseModel):
    # used by class Invoice
    value: str


class Parent(BaseModel):
    value: int
    link: str


class PatternType(BaseModel):
    value: str


class PaymentMethod(BaseModel):
    value: str


class PaymentStatus(BaseModel):
    value: str


class PhysicalCondition(BaseModel):
    value: str


class PhysicalMaterialType(BaseModel):
    value: str


class Policy(BaseModel):
    value: str


class Provenance(BaseModel):
    value: str


class RenewalCycle(BaseModel):
    value: Optional[str] = None


class ReportingCode(BaseModel):
    value: Optional[str] = None


class ReviewStatus(BaseModel):
    value: str
    desc: str


class SecondaryReportingCode(BaseModel):
    value: Optional[str] = None


class SourceType(BaseModel):
    value: Optional[str] = None


class Status1(BaseModel):
    # used by class Fund
    value: str
    desc: str


class Status2(BaseModel):
    # used by class License
    value: str
    desc: str


class TempCallNumberType(BaseModel):
    value: str


class TempLibrary(BaseModel):
    value: str


class TempLocation(BaseModel):
    value: str


class TempPolicy(BaseModel):
    value: str


class TertiaryReportingCode(BaseModel):
    value: Optional[str] = None


class Type1(BaseModel):
    # used by class Fund
    value: str
    desc: str


class Type2(BaseModel):
    # used by class PoLine
    value: str
    desc: str


class Type3(BaseModel):
    # used by class InvoiceVat
    value: str


class Type4(BaseModel):
    # used by class InvoiceLine
    value: str
    desc: str


class Type5(BaseModel):
    # used by class License
    value: str
    desc: str


class VatCode1(BaseModel):
    # used by class InvoiceVat
    value: str


class VatCode2(BaseModel):
    # used by class InvoiceLineVat
    value: str


class Vendor1(BaseModel):
    # used by class PoLine
    value: str
    desc: str


class Vendor2(BaseModel):
    # used by class Invoice
    value: str


class VendorReferenceNumberType(BaseModel):
    value: str


class VoucherCurrency(BaseModel):
    value: str


class Fund(BaseModel):
    id_: str = Field(..., alias='id')
    link: str
    code: str
    name: str
    external_id: Optional[str] = None
    type_: Optional[Type1] = Field(..., alias='type')
    entity_type: EntityType
    owner: Owner1
    status: Status1
    description: Optional[str] = None
    fiscal_period: FiscalPeriod
    currency: Currency1
    allocated_balance: float
    expended_balance: float
    cash_balance: float
    encumbered_balance: float
    available_balance: float
    available_for_libraries: List[AvailableForLibrary] = Field(..., alias='available_for_library')
    parent: Parent
    overencumbrance_allowed: OverencumbranceAllowed
    overexpenditure_allowed: OverexpenditureAllowed
    overencumbrance_warning_percent: int
    overexpenditure_warning_sum: float
    overencumbrance_limit_percent: int
    overexpenditure_limit_sum: float
    encumbrances_prior_to_fiscal_period: int
    expenditures_prior_to_fiscal_period: int
    transfers_prior_to_fiscal_period: int
    fiscal_period_end_encumbrance_grace_period: int
    fiscal_period_end_expenditure_grace_period: int


class Funds(BaseModel):
    total_record_count: int
    funds: List[Fund] = Field(..., alias='fund')


class Type6(BaseModel):
    # used by class FundTransaction
    desc: str
    value: str


class Currency5(BaseModel):
    # used by class FundTransaction
    desc: Optional[str] = None
    value: str


class PoLine2(BaseModel):
    # used by class FundTransaction
    link: str
    value: str


class InvoiceLine2(BaseModel):
    # used by class FundTransaction
    link: Optional[str] = None
    value: str


class FundTransaction(BaseModel):
    link: Optional[str] = None
    id_: str = Field(..., alias='id')
    transaction_time: date
    payment_time: Optional[str] = None
    type_: Type6 = Field(..., alias='type')
    amount: float
    currency: Currency5
    transaction_note: Optional[str] = None
    po_line: PoLine2
    invoice_line: InvoiceLine2
    reporting_code: str
    secondary_reporting_code: str
    tertiary_reporting_code: str
    fourth_reporting_code: str
    fifth_reporting_code: str

    _strip_gmt_date_z = validator('transaction_time', pre=True, allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class FundTransactions(BaseModel):
    total_record_count: int
    fund_transactions: List[FundTransaction] = Field(..., alias='fund_transaction')


class Alert(BaseModel):
    value: Optional[str] = None
    desc: str


class Currency2(BaseModel):
    # used by class Price
    value: str


class Currency4(BaseModel):
    # used by class Amount
    value: Optional[str] = None
    desc: Optional[str] = None


class PermanentLibrary(BaseModel):
    value: str


class Status3(BaseModel):
    # used by class PoLine
    value: str
    desc: str


class Price(BaseModel):
    sum_: float = Field(..., alias='sum')
    currency: Currency2


class Amount(BaseModel):
    sum_: Optional[float] = Field(..., alias='sum')
    currency: Currency4

    @validator('sum_', pre=True)
    def empty_string(value, field):
        if value == "":
            return None
        return value


class ResourceMetadata(BaseModel):
    mms_id: Optional[MmsId] = None
    title: Optional[str] = None
    author: Optional[str] = None
    issn: Optional[str] = None
    isbn: Optional[str] = None
    publisher: Optional[str] = None
    publication_place: Optional[str] = None
    publication_year: Optional[str] = None
    vendor_title_number: Optional[str] = None
    title_id: Optional[str] = None
    system_control_numbers: Optional[List[str]] = Field(alias='system_control_number')


class FundDistribution1(BaseModel):
    # used by class PoLine
    fund_code: FundCode1
    percent: Optional[float] = 0
    amount: Optional[Amount] = None


class Copy(BaseModel):
    link: str
    barcode: Optional[str] = None
    item_policy: ItemPolicy
    receive_date: Optional[str] = None
    enumeration_a: Optional[str] = None
    enumeration_b: Optional[str] = None
    enumeration_c: Optional[str] = None
    chronology_i: Optional[str] = None
    chronology_j: Optional[str] = None
    chronology_k: Optional[str] = None
    description: Optional[str] = None
    storage_location_id: Optional[str] = None
    is_temp_location: Optional[bool] = None
    permanent_library: Optional[PermanentLibrary] = None
    permanent_shelving_location: Optional[str] = None


class Location1(BaseModel):
    # used by class PoLine
    quantity: int
    library: Library
    shelving_location: str
    copies: List[Copy] = Field(..., alias='copy')


class PoLine(BaseModel):
    number: str
    status: Status3
    link: Optional[str] = None
    owner: Owner2
    type_: Type2 = Field(..., alias='type')
    vendor: Vendor1
    vendor_account: str
    reclaim_interval: Optional[str] = None
    expected_receipt_interval: Optional[str] = None
    claiming_interval: Optional[str] = None
    expected_activation_interval: Optional[str] = None
    subscription_interval: Optional[str] = None
    expected_activation_date: Optional[str] = None
    e_activation_due_interval: Optional[str] = None
    acquisition_method: Optional[AcquisitionMethod] = None
    no_charge: Optional[bool] = None
    rush: Optional[bool] = None
    cancellation_restriction: Optional[bool] = None
    cancellation_restriction_note: Optional[str] = None
    price: Optional[Price] = None
    discount: Optional[str] = None
    vendor_reference_number: str
    vendor_reference_number_type: Optional[VendorReferenceNumberType] = None
    source_type: SourceType
    po_number: str
    invoice_reference: Optional[str] = None
    resource_metadata: ResourceMetadata
    fund_distributions: Optional[List[FundDistribution1]] = Field(alias='fund_distribution')
    reporting_code: Optional[str] = None
    secondary_reporting_code: Optional[str] = None
    tertiary_reporting_code: Optional[str] = None
    fourth_reporting_code: Optional[str] = None
    fifth_reporting_code: Optional[str] = None
    vendor_note: Optional[str] = None
    receiving_note: Optional[str] = None
    alerts: Optional[List[Alert]] = Field(alias='alert')
    notes: Optional[List[Note1]] = Field(alias='note')
    locations: Optional[List[Location1]] = Field(alias='location')
    interested_users: Optional[List[InterestedUser]] = Field(alias='interested_user')
    license_: Optional[License2] = Field(..., alias='license')
    access_model: Optional[str] = None
    url: Optional[str] = None
    base_status: Optional[str] = None
    access_provider: Optional[str] = None
    manual_renewal: Optional[bool] = None
    renewal_cycle: Optional[RenewalCycle] = None
    subscription_from_date: Optional[date] = None
    subscription_to_date: Optional[date] = None
    renewal_date: Optional[date] = None
    renewal_period: Optional[str] = None
    renewal_note: Optional[str] = None
    material_type: Optional[MaterialType] = None
    expected_receipt_date: Optional[date] = None
    created_date: Optional[date] = None
    status_date: Optional[date] = None

    _strip_gmt_date_z = validator('expected_receipt_date',
                                  'created_date',
                                  'renewal_date',
                                  'status_date',
                                  'subscription_from_date',
                                  'subscription_to_date',
                                  pre=True,
                                  allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class PoLines(BaseModel):
    total_record_count: int
    po_lines: List[PoLine] = Field(..., alias='po_line')


class HoldingData(BaseModel):
    link: str
    holding_id: str
    copy_id: str
    in_temp_location: bool
    temp_library: TempLibrary
    temp_location: TempLocation
    temp_call_number_type: TempCallNumberType
    temp_call_number: str
    temp_call_number_source: str
    temp_policy: TempPolicy
    due_back_date: str


class ItemData(BaseModel):
    barcode: str
    physical_material_type: PhysicalMaterialType
    policy: Policy
    provenance: Provenance
    po_line: str
    issue_date: str
    is_magnetic: bool
    arrival_date: str
    expected_arrival_date: str
    year_of_issue: str
    enumeration_a: str
    enumeration_b: str
    enumeration_c: str
    enumeration_d: str
    enumeration_e: str
    enumeration_f: str
    enumeration_g: str
    enumeration_h: str
    chronology_i: str
    chronology_j: str
    chronology_k: str
    chronology_l: str
    chronology_m: str
    break_indicator: BreakIndicator
    pattern_type: PatternType
    linking_number: str
    description: str
    replacement_cost: int
    receiving_operator: str
    inventory_number: str
    inventory_date: str
    inventory_price: str
    receive_number: str
    weeding_number: str
    weeding_date: str
    alternative_call_number: str
    alternative_call_number_type: AlternativeCallNumberType
    alt_number_source: str
    storage_location_id: str
    pages: str
    pieces: str
    public_note: str
    fulfillment_note: str
    internal_note_1: str
    internal_note_2: str
    internal_note_3: str
    statistics_note_1: str
    statistics_note_2: str
    statistics_note_3: str
    physical_condition: PhysicalCondition


class Item(BaseModel):
    link: str
    holding_data: HoldingData
    item_data: ItemData


class PoLineItems(BaseModel):
    total_record_count: int
    items: List[Item] = Field(..., alias='item')


class InvoiceVat(BaseModel):
    report_tax: bool
    vat_per_invoice_line: bool
    vat_code: VatCode1
    percentage: int
    vat_amount: float
    type_: Type3 = Field(..., alias='type')
    expended_from_fund: bool
    vendor_tax: Optional[str] = None


class Payment(BaseModel):
    prepaid: bool
    internal_copy: bool
    export_to_erp: Optional[bool] = None
    payment_status: PaymentStatus
    voucher_date: str
    voucher_number: str
    voucher_amount: str
    voucher_currency: VoucherCurrency


class InvoiceLineVat(BaseModel):
    vat_code: VatCode2
    percentage: int
    vat_amount: int


class FundDistribution2(BaseModel):
    # used by class InvoiceLine
    fund_code: FundCode2
    percent: float
    amount: float


class InvoiceLine(BaseModel):
    link: str
    type_: Type4 = Field(..., alias='type')
    number: str
    po_line: str
    price: float
    price_note: str
    quantity: int
    vat_note: str
    check_subscription_date_overlap: bool
    fully_invoiced: bool
    subscription_from_date: Optional[date] = None
    subscription_to_date: Optional[date] = None
    additional_info: Optional[str] = None
    release_remaining_encumbrance: bool
    reporting_code: Optional[ReportingCode] = None
    secondary_reporting_code: Optional[SecondaryReportingCode] = None
    tertiary_reporting_code: Optional[TertiaryReportingCode] = None
    note: str
    invoice_line_vat: InvoiceLineVat
    fund_distributions: List[FundDistribution2] = Field(..., alias='fund_distribution')

    _strip_gmt_date_z = validator('subscription_from_date',
                                  'subscription_to_date',
                                  pre=True,
                                  allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class InvoiceLines(BaseModel):
    total_record_count: int
    invoice_lines: List[InvoiceLine] = Field(..., alias='invoice_line')


class ExplicitRatio(BaseModel):
    foreign_currency: ForeignCurrency
    rate: int


class Invoice(BaseModel):
    link: str
    number: str
    invoice_date: date
    invoice_due_date: Optional[date] = None
    vendor: Vendor2
    vendor_account: str
    total_amount: int
    currency: Currency3
    payment_method: PaymentMethod
    reference_number: str
    owner: Owner3
    additional_charges: AdditionalCharges
    invoice_vat: InvoiceVat
    # plural in 's' used for clarity instead of 'ratii'
    explicit_ratios: List[ExplicitRatio] = Field(..., alias='explicit_ratio')
    payment: Payment
    notes: List[Note2] = Field(..., alias='note')
    invoice_lines: InvoiceLines

    _strip_gmt_date_z = validator('invoice_date',
                                  'invoice_due_date',
                                  pre=True,
                                  allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class Invoices(BaseModel):
    total_record_count: int
    invoices: List[Invoice] = Field(..., alias='invoice')


class Amendments(BaseModel):
    value: int
    link: str


class Code(BaseModel):
    value: str
    desc: Optional[str] = None


class Value(BaseModel):
    value: str
    desc: Optional[str] = None


class Term(BaseModel):
    code: Code
    value: Value


class Type6(BaseModel):
    # used by class Resource
    value: str
    desc: str


class Resource(BaseModel):
    pid: str
    name: str
    type_: Type6 = Field(..., alias='type')
    link: str


class Type7(BaseModel):
    # used by class Note3
    value: str
    desc: str


class Note3(BaseModel):
    # used by class License
    content: Optional[str] = None
    creation_date: date
    created_by: str
    type_: Optional[Type7] = Field(..., alias='type')

    _strip_gmt_date_z = validator('creation_date', pre=True, allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class License(BaseModel):
    link: str
    code: str
    name: str
    type_: Type5 = Field(..., alias='type')
    status: Status2
    licensor: Licensor
    signed_by: Optional[str] = None
    signed_date: Optional[date] = None
    second_party_signed_by: Optional[str] = None
    second_party_signed_date: Optional[str] = None
    start_date: date
    end_date: Optional[date] = None
    location: Optional[Location2] = None
    review_status: ReviewStatus
    uri: Optional[str] = None
    amendments: Optional[Amendments] = None
    licensing_agent: Optional[LicensingAgent] = None
    terms: Optional[List[Term]] = Field(alias='term')
    resources: Optional[List[Resource]] = Field(alias='resource')
    notes: Optional[List[Note3]] = Field(alias='note')
    administrators: Optional[List[Administrator]] = Field(alias='administrator')

    _strip_gmt_date_z = validator('end_date', 'signed_date', 'start_date', pre=True, allow_reuse=True)(
        common_validators._strip_gmt_date_z
    )


class Licenses(BaseModel):
    total_record_count: int
    licenses: List[License] = Field(..., alias='license')



# Copyright 2022 University of Waterloo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import urljoin

import json
import logging

from almonaut.session import AlmaApiSession
from almonaut.exceptions import handle_error_response

from almonaut.acquisitions import acquisitions_models
from almonaut.electronic_resources import electronic_resources_models

logging.basicConfig(level=logging.DEBUG)
# logging.basicConfig(level=logging.WARNING)


class AlmaApiClient(object):
    """The Alma API client.

    :param api_key: The Alma API key to use.
    :param host: Hostname of the Alma API instance.
    :param url_prefix: Prefix before the API version.
    :param version: API version to use.
    """

    def __init__(self,
                 api_key: str,
                 host: str = 'https://api-ca.hosted.exlibrisgroup.com',
                 url_prefix: str = 'almaws',
                 version: str = 'v1'):
        """Instantiate a new API client."""
        self.api_key = api_key
        self.host = host
        self.url_prefix = url_prefix
        self.version = version
        self.session = AlmaApiSession()

    def _request(self, limit=5, offset=0):
        """Execute an API request."""
        rel_url = "/".join((self.url_prefix, self.version, self.end_point))
        target_url = urljoin(self.host, rel_url)
        params = {'apikey': self.api_key, 'format': self.format_,
                  'limit': limit, 'offset': offset}
        params = {**params, **self.extra_params}
        response = self.session.request(self.method, target_url,
                                        params=params)
        logging.info("************* API hit ***************")
        logging.debug(response.url)
        if response.status_code >= 400:
            handle_error_response(response)
        else:
            return response

    def _get_records(self, end_point=None, format_='json',
                     limit=5, all_records=False, extra_params=None,
                     data_dict_key=None):
        """Retrieve records for a query.

        If the number of records for the query exceeds the limit, make multiple
        API calls until all records for the query are retrieved.
        """
        self.end_point = end_point
        self.format_ = format_
        self.extra_params = extra_params
        self.method = 'GET'

        response = self._request(limit=limit, offset=0)
        if type(json.loads(response.content)) == dict:
            response_json = json.loads(response.content)
            if 'total_record_count' in response_json:
                total_records = response_json['total_record_count']
            else:
                total_records = 1

        if total_records == 0:
            return
        elif not all_records:
            return response.content
        else:
            records_requested = limit
            limit = 50
            while records_requested < total_records:
                subsequent_response = self._request(limit=limit,
                                                    offset=records_requested)
                records_requested += limit
                if type(json.loads(subsequent_response.content)) == dict:
                    subsequent_response_json = json.loads(subsequent_response.content)
                    response_json[data_dict_key] += subsequent_response_json[data_dict_key]
            back_to_str = json.dumps(response_json)

            return back_to_str

    # API methods

    # acquisitions

    def get_fund(self, id_: str, format_: str = 'json') -> acquisitions_models.Fund:
        r"""Get a fund record.

        :param id\_: Alma fund ID.
        :param format\_: Format of the raw returned data.

        The ``extra_params`` dict can include:

        view
          brief|full
        """
        end_point = f"acq/funds/{id_}"
        extra_params = {'view': 'full'}
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=1,
                                   all_records=False,
                                   extra_params=extra_params
                                   )
        if result:
            logging.debug(result)
            return acquisitions_models.Fund.parse_raw(result)

    def get_funds(self, format_: str = 'json', limit: int = 5,
                  all_records: bool = False, extra_params={}) -> acquisitions_models.Funds:
        r"""Get fund records.

        :param format\_: Format of the raw returned data.
        :param limit: The maximum number of records to be returned.
        :param all_records: Whether or not all matching records should be returned (overrides ``limit``).
        :param extra_params: Additional parameters.

        The ``extra_params`` dict can include:

        view
          brief|full
        """
        extra_params['view'] = 'full'
        result = self._get_records(end_point='acq/funds',
                                   format_=format_,
                                   limit=limit,
                                   all_records=all_records,
                                   extra_params=extra_params,
                                   data_dict_key='fund'
                                   )
        if result:
            logging.debug(result)
            return acquisitions_models.Funds.parse_raw(result)

    def get_fund_transactions(self, fund_id: str, format_: str = 'json',
                              limit: int = 5, all_records: bool = False,
                              extra_params={}) -> acquisitions_models.FundTransactions:
        r"""Get fund transaction records.

        :param fund_id: Alma fund ID.
        :param format\_: Format of the raw returned data.
        :param limit: The maximum number of records to be returned.
        :param all_records: Whether or not all matching records should be returned (overrides ``limit``).
        :param extra_params: Additional parameters.
        """
        end_point = f"acq/funds/{fund_id}/transactions"
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=limit,
                                   all_records=all_records,
                                   extra_params=extra_params,
                                   data_dict_key='fund_transaction'
                                   )
        if result:
            logging.debug(result)
            return acquisitions_models.FundTransactions.parse_raw(result)

    def get_invoice(self, invoice_id: str, format_: str = 'json') -> acquisitions_models.Invoice:
        r"""Get an invoice record.

        :param invoice_id: Alma invoice ID.
        :param format\_: Format of the raw returned data.

        The ``extra_params`` dict can include:

        expand
          attachments|none
        view
          brief|full
        """
        end_point = f"acq/invoices/{invoice_id}"
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=1,
                                   all_records=False,
                                   extra_params={}
                                   )
        if result:
            logging.debug(result)
            return acquisitions_models.Invoice.parse_raw(result)

    def get_invoices(self, format_: str = 'json', limit: int = 5,
                     all_records: bool = False, extra_params={}) -> acquisitions_models.Invoices:
        r"""Get invoice records.

        :param format\_: Format of the raw returned data.
        :param limit: The maximum number of records to be returned.
        :param all_records: Whether or not all matching records should be returned (overrides ``limit``).
        :param extra_params: Additional parameters.

        The ``extra_params`` dict can include:

        base_status
          ACTIVE|All|CLOSED
        creation_form
          (various)
        expand
          attachments|none
        invoice_workflow_status
          (various)
        owner
          (invoice owner: can be the institution code or a library code)
        view
          brief|full
        """
        end_point = 'acq/invoices/'
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=limit,
                                   all_records=all_records,
                                   extra_params=extra_params,
                                   data_dict_key='invoice'
                                   )
        if result:
            logging.debug(result)
            return acquisitions_models.Invoices.parse_raw(result)

    def get_invoice_line(self, invoice_id: str, invoice_line_id: str,
                         format_: str = 'json') -> acquisitions_models.InvoiceLine:
        r"""Get an invoice line record.

        :param invoice_id: Alma invoice ID.
        :param invoice_line_id: Alma invoice line ID.
        :param format\_: Format of the raw returned data.
        """
        end_point = f"acq/invoices/{invoice_id}/lines/{invoice_line_id}"
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=1,
                                   all_records=False,
                                   extra_params={}
                                   )
        if result:
            logging.debug(result)
            return acquisitions_models.InvoiceLine.parse_raw(result)

    def get_invoice_lines(self, invoice_id: str, format_: str = 'json',
                          limit: int = 5, all_records: bool = False,
                          extra_params={}) -> acquisitions_models.InvoiceLines:
        r"""Get invoice line records.

        :param invoice_id: Alma invoice ID.
        :param format\_: Format of the raw returned data.
        :param limit: The maximum number of records to be returned.
        :param all_records: Whether or not all matching records should be returned (overrides ``limit``).
        :param extra_params: Additional parameters.
        """
        end_point = f"acq/invoices/{invoice_id}/lines"
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=limit,
                                   all_records=all_records,
                                   extra_params=extra_params,
                                   data_dict_key='invoice_line'
                                   )
        if result:
            logging.debug(result)
            return acquisitions_models.InvoiceLines.parse_raw(result)

    def get_license(self, code: str, format_: str = 'json') -> acquisitions_models.License:
        r"""Get a license record.

        :param code: Alma license code.
        :param format\_: Format of the raw returned data.

        The ``extra_params`` dict can include:

        expand
          attachments|none
        include_blank_terms
          true|false
        """
        end_point = f"acq/licenses/{code}"
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=1,
                                   all_records=False,
                                   extra_params={}
                                   )
        if result:
            logging.debug(result)
            return acquisitions_models.License.parse_raw(result)

    def get_licenses(self, format_: str = 'json', limit: int = 5,
                     all_records: bool = False, extra_params={}) -> acquisitions_models.Licenses:
        r"""Get license records.

        :param format\_: Format of the raw returned data.
        :param limit: The maximum number of records to be returned.
        :param all_records: Whether or not all matching records should be returned (overrides ``limit``).
        :param extra_params: Additional parameters.
        """
        # extra_params['expand'] = 'attachments'
        end_point = 'acq/licenses'
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=limit,
                                   all_records=all_records,
                                   extra_params=extra_params,
                                   data_dict_key='license'
                                   )
        if result:
            logging.debug(result)
            return acquisitions_models.Licenses.parse_raw(result)

    def get_po_line(self, number: str, format_: str = 'json') -> acquisitions_models.PoLine:
        r"""Get a PO Line record.

        :param number: Alma PO Line number.
        :param format\_: Format of the raw returned data.
        """
        end_point = f"acq/po-lines/{number}"
        # extra_params['expand'] = 'notes'
        # extra_params['expand'] = 'locations, notes'
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=1,
                                   all_records=False,
                                   extra_params={},
                                   )
        if result:
            logging.debug(result)
            return acquisitions_models.PoLine.parse_raw(result)

    def get_po_lines(self, format_: str = 'json', limit: int = 5,
                     all_records: bool = False, extra_params={}) -> acquisitions_models.PoLines:
        r"""Get PO Line records.

        :param format\_: Format of the raw returned data.
        :param limit: The maximum number of records to be returned.
        :param all_records: Whether or not all matching records should be returned (overrides ``limit``).
        :param extra_params: Additional parameters.

        The ``extra_params`` dict can include:

        acquisition_method
          PURCHASE|ALL
        expand
          locations|notes|locations, notes
        status
          ACTIVE|ALL|ALL_WITH_CLOSED|CANCELLED|CLOSED
        q
          *e.g.:* number~123456, po_number~PO123, title~spenser
        """
        end_point = 'acq/po-lines'
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=limit,
                                   all_records=all_records,
                                   extra_params=extra_params,
                                   data_dict_key='po_line'
                                   )
        if result:
            logging.debug(result)
            return acquisitions_models.PoLines.parse_raw(result)

    # e-resources

    def get_electronic_collection(self, collection_id: str,
                                  format_: str = 'json') -> electronic_resources_models.ElectronicCollection:
        r"""Get an Electronic Collection record.

        :param collection_id: Alma electronic collection ID.
        :param format\_: Format of the raw returned data.
        """
        result = self._get_records(end_point=f"electronic/e-collections/{collection_id}",
                                   format_=format_,
                                   limit=1,
                                   all_records=False,
                                   extra_params={}
                                   )
        if result:
            logging.debug(result)
            return electronic_resources_models.ElectronicCollection.parse_raw(result)

    def get_electronic_collections(self, format_: str = 'json', limit: int = 5,
                                   all_records: bool = False, extra_params={}) -> electronic_resources_models.ElectronicCollections:
        r"""Get Electronic Collection records.

        :param format\_: Format of the raw returned data.
        :param limit: The maximum number of records to be returned.
        :param all_records: Whether or not all matching records should be returned (overrides ``limit``).
        :param extra_params: Additional parameters.
        """
        result = self._get_records(end_point='electronic/e-collections',
                                   format_=format_,
                                   limit=limit,
                                   all_records=all_records,
                                   extra_params=extra_params,
                                   data_dict_key='electronic_collection'
                                   )
        if result:
            logging.debug(result)
            return electronic_resources_models.ElectronicCollections.parse_raw(result)

    def get_electronic_service(self, collection_id: str, service_id: str,
                               format_: str = 'json') -> electronic_resources_models.ElectronicService:
        r"""Get an Electronic Service record.

        :param collection_id: Alma electronic collection ID.
        :param service_id: Alma electronic service ID.
        :param format\_: Format of the raw returned data.
        """
        end_point = f"electronic/e-collections/{collection_id}/e-services/{service_id}"
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=1,
                                   all_records=False,
                                   extra_params={}
                                   )
        if result:
            logging.debug(result)
            return electronic_resources_models.ElectronicService.parse_raw(result)

    def get_electronic_services(self, collection_id: str,
                                format_: str = 'json', limit: int = 5,
                                all_records: bool = False, extra_params={}) -> electronic_resources_models.ElectronicServices:
        r"""Get Electronic Service records.

        :param collection_id: Alma electronic collection ID.
        :param format\_: Format of the raw returned data.
        :param limit: The maximum number of records to be returned.
        :param all_records: Whether or not all matching records should be returned (overrides ``limit``).
        :param extra_params: Additional parameters.
        """
        end_point = f"electronic/e-collections/{collection_id}/e-services"
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=limit,
                                   all_records=all_records,
                                   extra_params=extra_params,
                                   data_dict_key='electronic_service'
                                   )
        if result:
            logging.debug(result)
            return electronic_resources_models.ElectronicServices.parse_raw(result)

    def get_portfolio(self, collection_id: str, service_id: str,
                      portfolio_id: str, format_: str = 'json') -> electronic_resources_models.Portfolio:
        r"""Get a Portfolio record.

        :param collection_id: Alma electronic collection ID.
        :param service_id: Alma electronic service ID.
        :param portfolio_id: Alma portfolio ID.
        :param format\_: Format of the raw returned data.
        """
        end_point = f"electronic/e-collections/{collection_id}/e-services/{service_id}/portfolios/{portfolio_id}"
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=1,
                                   all_records=False,
                                   extra_params={}
                                   )
        if result:
            logging.debug(result)
            return electronic_resources_models.Portfolio.parse_raw(result)

    def get_portfolios(self, collection_id: str, service_id: str,
                       format_: str = 'json', limit: int = 5,
                       all_records: bool = False, extra_params={}) -> electronic_resources_models.Portfolios:
        r"""Get Portfolio records.

        :param collection_id: Alma electronic collection ID.
        :param service_id: Alma electronic service ID.
        :param format\_: Format of the raw returned data.
        :param limit: The maximum number of records to be returned.
        :param all_records: Whether or not all matching records should be returned (overrides ``limit``).
        :param extra_params: Additional parameters.
        """
        end_point = f"electronic/e-collections/{collection_id}/e-services/{service_id}/portfolios"
        result = self._get_records(end_point=end_point,
                                   format_=format_,
                                   limit=limit,
                                   all_records=all_records,
                                   extra_params=extra_params,
                                   data_dict_key='portfolio'
                                   )
        if result:
            logging.debug(result)
            return electronic_resources_models.Portfolios.parse_raw(result)



# Copyright 2022 University of Waterloo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def handle_error_response(resp):
    """Handle Alma API exceptions."""
    codes = {
        -1: AlmaApiError,
        -402119: GeneralError,
        -40166410: InvalidParameterWithValidOptionsError,
        -40166419: NoValidOptionsParameterError,
        -401873: NoFilterWithPolModeError,
    }
    error = resp.json().get('error', {})
    message = error.get('message')
    code = error.get('code', -1)
    data = error.get('data', {})
    raise codes[code](message=message, code=code, data=data, response=resp)


class AlmaApiError(Exception):
    """An Alma API exception."""

    response = None
    data = {}
    code = -1
    message = "An unknown error occurred"

    def __init__(self, message=None, code=None, data={}, response=None):
        """Init method."""
        self.response = response
        if message:
            self.message = message
        if code:
            self.code = code
        if data:
            self.data = data

    def __str__(self):
        """Create string representation."""
        if self.code:
            return '{}: {}'.format(self.code, self.message)
        return self.message


class GeneralError(AlmaApiError):
    """Handle an Alma API GeneralError exception."""

    pass


class InvalidParameterWithValidOptionsError(AlmaApiError):
    """Handle an Alma API InvalidParameterWithValidOptionsError exception."""

    pass


class NoValidOptionsParameterError(AlmaApiError):
    """Handle an Alma API NoValidOptionsParameterError exception."""

    pass


class NoFilterWithPolModeError(AlmaApiError):
    """Handle an Alma API NoFilterWithPolModeError exception."""

    pass
